import docker


def get_hostname():
    # mmmh, how should this be done? need nome kind ouf routing...
    return 'not imlemented yet'


def _docker_client():
    return docker.DockerClient(base_url='unix://var/run/docker.sock', version='auto', timeout=30)


def get_open_port():
    # this should be done somewhere else, e.g docker itself, but for now...
    import socket
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    s.listen(1)
    port = s.getsockname()[1]
    s.close()
    return port


def get_container_by_name(postgraas_instance_name):
    # return the container object or None if not found
    c = _docker_client()
    try:
        return c.containers.get(postgraas_instance_name)
    except docker.errors.NotFound:
        return None


def check_container_exists(postgraas_instance_name):
    if get_container_by_name(postgraas_instance_name):
        return True
    return False


def create_postgres_instance(postgraas_instance_name, connection_dict):
    c = _docker_client()
    environment = {
        "POSTGRES_USER": connection_dict['db_username'],
        "POSTGRES_PASSWORD": connection_dict['db_pwd'],
        "POSTGRES_DB": connection_dict['db_name']
    }
    internal_port = 5432
    if check_container_exists(postgraas_instance_name):
        raise ValueError('Container exists already')
    image = 'postgres:9.4'
    container = c.containers.create(
        image,
        name=postgraas_instance_name,
        ports={internal_port: connection_dict['port']},
        environment=environment,
        restart_policy={"Name": "unless-stopped"},
        labels={"postgraas": image}
    )
    container.start()
    return container.id


def delete_postgres_instance(container_id):
    c = _docker_client()
    c.containers.get(container_id).remove(force=True)



import ConfigParser

from .docker import DockerBackend
from .postgres_cluster import PGClusterBackend

BACKENDS = {'docker': DockerBackend, 'pg_cluster': PGClusterBackend}


def get_backend(config):
    backend_config = {}
    try:
        backend_config = dict(config.items('backend'))
        backend = backend_config['type']
    except ConfigParser.NoSectionError:
        backend = 'docker'
    return BACKENDS[backend](backend_config)



import os

from cryptography.fernet import Fernet

import postgraas_server.configuration as cf


class TestConfiguration:
    module_path = os.path.abspath(os.path.dirname(__file__))

    def test_default_config_filename(self):
        actual = cf.get_default_config_filename()
        expected = os.path.join(os.getcwd(), 'application.cfg')
        assert actual == expected

    def test_get_config(self):
        test_config = os.path.join(self.module_path, 'application.cfg')
        actual = cf.get_config(test_config)
        expected = 'postgraas'
        assert actual.get('metadb', 'db_name') == expected

    def test_secrets(self, tmpdir):
        secrets_file = tmpdir.join('secrets')
        key = Fernet.generate_key()
        print(key)
        f = Fernet(key)
        secrets_file.write('{{"encryption_key": "{}"}}'.format(key))
        test_config = os.path.join(self.module_path, 'application.cfg')
        cfg_file = tmpdir.join('application.cfg')
        cfg_file.write(f.encrypt(open(test_config, 'rb').read()))
        config = cf.get_config(cfg_file.strpath, secrets_file=secrets_file.strpath)
        assert config.get('metadb', 'db_name') == 'postgraas'



import os
import json
import uuid

import docker
import pytest
from mock import patch, MagicMock, Mock

import postgraas_server.backends.docker.postgres_instance_driver as pid
import postgraas_server.backends.postgres_cluster.postgres_cluster_driver as pgcd
import postgraas_server.configuration as configuration
from postgraas_server.backends.exceptions import PostgraasApiException
from postgraas_server.create_app import create_app
from .utils import wait_for_postgres_listening

DOCKER_CONFIG = """
[metadb]
db_name = postgraas
db_username = postgraas
db_pwd = postgraas12
host = localhost
port = 54321

[backend]
type = docker
"""

CLUSTER_CONFIG = """
[metadb]
db_name = postgraas
db_username = postgraas
db_pwd = postgraas12
host = localhost
port = 54321

[backend]
type = pg_cluster
host = {host}
port = {port}
database = {database}
username = {username}
password = {password}
""".format(
    database=os.environ.get('PGDATABASE', 'postgres'),
    username=os.environ.get('PGUSER', 'postgres'),
    password=os.environ.get('PGPASSWORD', 'postgres'),
    port=os.environ.get('PGPORT', '5432'),
    host=os.environ.get('PGHOST', 'localhost')
)

CONFIGS = {
    'docker': DOCKER_CONFIG,
    'pg_cluster': CLUSTER_CONFIG,
}


def remove_digits(s):
    return ''.join(c for c in s if not c.isdigit())


def delete_all_test_postgraas_container():
    c = pid._docker_client()
    for container in c.containers.list():
        if container.name.startswith("tests_postgraas_"):
            container.remove(force=True)


def delete_test_database_and_user(db_name, username, config):
    pgcd.delete_database(db_name, config)
    pgcd.delete_user(username, config)


@pytest.fixture(params=['docker', 'pg_cluster'])
def parametrized_setup(request, tmpdir):
    from postgraas_server.management_resources import db
    cfg = tmpdir.join('config')
    cfg.write(CONFIGS[request.param])
    config = configuration.get_config(cfg.strpath)
    this_app = create_app(config)
    this_app.config['SQLALCHEMY_DATABASE_URI'] = "sqlite://"
    this_app.use_reloader = False
    this_app.config['TESTING'] = True
    ctx = this_app.app_context()
    ctx.push()
    db.create_all()
    username, db_name = str(uuid.uuid4()).replace('-', '_'), str(uuid.uuid4()).replace('-', '_')
    request.cls.app_client = this_app.test_client()
    request.cls.db_name = remove_digits(db_name)
    request.cls.username = remove_digits(username)
    request.cls.backend = request.param
    yield
    if request.param == 'docker':
        delete_all_test_postgraas_container()
    elif request.param == 'pg_cluster':
        try:
            delete_test_database_and_user(db_name, username, dict(config.items('backend')))
        except Exception:
            pass
    db.drop_all()
    ctx.pop()


@pytest.fixture()
def docker_setup(request, tmpdir):
    from postgraas_server.management_resources import db
    cfg = tmpdir.join('config')
    cfg.write(CONFIGS['docker'])
    this_app = create_app(configuration.get_config(cfg.strpath))
    this_app.config['SQLALCHEMY_DATABASE_URI'] = "sqlite://"
    this_app.use_reloader = False
    this_app.config['TESTING'] = True
    ctx = this_app.app_context()
    ctx.push()
    db.create_all()
    request.cls.app_client = this_app.test_client()
    yield
    delete_all_test_postgraas_container()
    db.drop_all()
    ctx.pop()


class PostgraasApiTestBase:
    def get_postgraas_by_name(self, name, client):
        headers = {'Content-Type': 'application/json'}
        list = client.get('/api/v2/postgraas_instances', headers=headers)
        for instance in json.loads(list.data):
            if instance["postgraas_instance_name"] == name:
                return instance["id"]

    def delete_instance_by_name(self, db_credentials, client):
        id = self.get_postgraas_by_name(db_credentials["postgraas_instance_name"], client)
        db_pwd = db_credentials["db_pwd"]
        headers = {'Content-Type': 'application/json'}
        client.delete(
            '/api/v2/postgraas_instances/' + str(id),
            data=json.dumps({
                'db_pwd': db_pwd
            }),
            headers=headers
        )


@pytest.mark.usefixtures('docker_setup')
class TestPostgraasApiDocker(PostgraasApiTestBase):
    def test_create_postgres_instance_api(self):
        db_credentials = {
            "postgraas_instance_name": "tests_postgraas_test_create_postgres_instance_api",
            "db_name": "test_create_postgres_instance",
            "db_username": "db_user",
            "db_pwd": "secret"
        }
        self.delete_instance_by_name(db_credentials, self.app_client)
        headers = {'Content-Type': 'application/json'}
        result = self.app_client.post(
            '/api/v2/postgraas_instances', headers=headers, data=json.dumps(db_credentials)
        )
        created_db = json.loads(result.data)
        assert created_db["db_name"] == 'test_create_postgres_instance'
        self.delete_instance_by_name(db_credentials, self.app_client)

    def test_create_docker_fails(self):
        db_credentials = {
            "postgraas_instance_name": "tests_postgraas_test_create_postgres_instance_api",
            "db_name": "test_create_postgres_instance",
            "db_username": "db_user",
            "db_pwd": "secret"
        }
        self.delete_instance_by_name(db_credentials, self.app_client)
        headers = {'Content-Type': 'application/json'}

        def raise_apierror(*args, **kwargs):
            raise PostgraasApiException('let create fail')

        with patch.object(docker.models.containers.ContainerCollection, 'create', raise_apierror):
            result = self.app_client.post(
                '/api/v2/postgraas_instances', headers=headers, data=json.dumps(db_credentials)
            )
        created_db = json.loads(result.data)
        assert 'let create fail' in created_db[
            "msg"
        ], 'unexpected error message for docker create failure'

    def test_delete_docker_notfound(self):
        db_credentials = {
            "postgraas_instance_name": "tests_postgraas_test_delete_docker_notfound",
            "db_name": "test_create_postgres_instance",
            "db_username": "db_user",
            "db_pwd": "secret"
        }
        self.delete_instance_by_name(db_credentials, self.app_client)
        headers = {'Content-Type': 'application/json'}
        result = self.app_client.post(
            '/api/v2/postgraas_instances', headers=headers, data=json.dumps(db_credentials)
        )
        created_db = json.loads(result.data)
        wait_success = wait_for_postgres_listening(created_db['container_id'])
        assert wait_success is True, 'postgres did not come up within 10s (or unexpected docker image log output)'

        def raise_not_found(*args, **kwargs):
            raise docker.errors.NotFound('raise for testing from mock')

        with patch.object(docker.models.containers.ContainerCollection, 'get', raise_not_found):
            res = self.app_client.delete(
                '/api/v2/postgraas_instances/' + str(created_db["postgraas_instance_id"]),
                data=json.dumps({
                    'db_pwd': db_credentials['db_pwd']
                }),
                headers=headers
            )
            res = json.loads(res.data)
        assert res['status'] == 'success'
        assert 'deleted postgraas instance, but container was not found' in res['msg']

    def test_driver_name_exists(self):
        db_credentials = {
            "db_name": 'test_db_name',
            "db_username": 'test_db_username',
            "db_pwd": 'test_db_pwd',
            "host": pid.get_hostname(),
            "port": pid.get_open_port()
        }
        if pid.check_container_exists('test_instance_name'):
            pid.delete_postgres_instance('test_instance_name')
        id0 = pid.create_postgres_instance('test_instance_name', db_credentials)
        with pytest.raises(ValueError):
            pid.create_postgres_instance('test_instance_name', db_credentials)

        pid.delete_postgres_instance(id0)
        assert pid.check_container_exists(id0) is False, "container exists after it was deleted"


@pytest.mark.usefixtures('parametrized_setup')
class TestPostgraasApi(PostgraasApiTestBase):
    def test_create_postgres_instance(self):
        db_credentials = {
            "db_name": 'test_db_name',
            "db_username": 'test_db_username',
            "db_pwd": 'test_db_pwd',
            "host": pid.get_hostname(),
            "port": pid.get_open_port()
        }
        mock_c = MagicMock()
        mock_c.id = 'fy8rfsufusgsufbvluluivhhvsbr'
        mock_create = Mock(return_value=mock_c)
        with patch.object(docker.models.containers.ContainerCollection, 'create', mock_create):
            result = pid.create_postgres_instance(
                'tests_postgraas_test_instance_name', db_credentials
            )
        assert result == 'fy8rfsufusgsufbvluluivhhvsbr'

    def test_delete_postgres_instance_api(self):
        db_credentials = {
            "postgraas_instance_name": "tests_postgraas_test_delete_postgres_instance_api",
            "db_name": self.db_name.decode(),
            "db_username": self.username.decode(),
            "db_pwd": "secret"
        }
        self.delete_instance_by_name(db_credentials, self.app_client)
        headers = {'Content-Type': 'application/json'}
        result = self.app_client.post(
            '/api/v2/postgraas_instances', headers=headers, data=json.dumps(db_credentials)
        )
        created_db = json.loads(result.data)

        if self.backend == 'docker':
            wait_success = wait_for_postgres_listening(created_db['container_id'])
            assert wait_success is True, 'postgres did not come up within 10s (or unexpected docker image log output)'

        delete_result = self.app_client.delete(
            '/api/v2/postgraas_instances/' + str(created_db["postgraas_instance_id"]),
            data=json.dumps({
                'db_pwd': 'wrong_password'
            }),
            headers=headers
        )
        deleted_db = json.loads(delete_result.data)

        assert deleted_db["status"] == 'failed'
        assert 'password authentication failed' in deleted_db[
            'msg'
        ], 'unexpected message for wrong password'

        def raise_apierror(*args, **kwargs):
            raise PostgraasApiException('let remove fail')

        with patch.object(docker.models.containers.Container, 'remove', raise_apierror):
            with patch.object(pgcd, 'delete_database', raise_apierror):
                delete_result = self.app_client.delete(
                    '/api/v2/postgraas_instances/' + str(created_db["postgraas_instance_id"]),
                    data=json.dumps({
                        'db_pwd': db_credentials['db_pwd']
                    }),
                    headers=headers
                )
                deleted_db = json.loads(delete_result.data)
                assert deleted_db["status"] == 'failed'
                assert 'let remove fail' in deleted_db[
                    'msg'
                ], 'unexpected error message on docker rm failure'

        delete_result = self.app_client.delete(
            '/api/v2/postgraas_instances/' + str(created_db["postgraas_instance_id"]),
            data=json.dumps({
                'db_pwd': db_credentials['db_pwd']
            }),
            headers=headers
        )
        deleted_db = json.loads(delete_result.data)
        assert deleted_db["status"] == 'success'

    def test_delete_notfound(self):
        headers = {'Content-Type': 'application/json'}
        res = self.app_client.delete(
            '/api/v2/postgraas_instances/123456789',
            data=json.dumps({
                'db_pwd': '123'
            }),
            headers=headers
        )
        res = json.loads(res.data)
        assert res['status'] == 'failed'
        assert "123456789" in res['msg'], 'unexpected error message'

    def test_create_postgres_instance_name_exists(self):
        db_credentials = {
            "postgraas_instance_name": "tests_postgraas_my_postgraas_twice",
            "db_name": self.db_name.decode(),
            "db_username": self.username.decode(),
            "db_pwd": "secret"
        }
        self.delete_instance_by_name(db_credentials, self.app_client)
        headers = {'Content-Type': 'application/json'}
        self.app_client.post(
            '/api/v2/postgraas_instances', headers=headers, data=json.dumps(db_credentials)
        )
        second = self.app_client.post(
            '/api/v2/postgraas_instances', headers=headers, data=json.dumps(db_credentials)
        )
        assert second.data == json.dumps(
            {
                "msg": "postgraas_instance_name already exists tests_postgraas_my_postgraas_twice"
            }
        ) + "\n"

        self.delete_instance_by_name(db_credentials, self.app_client)

    def test_return_postgres_instance_api(self):
        db_credentials = {
            u"postgraas_instance_name": u"tests_postgraas_test_return_postgres_instance_api",
            u"db_name": self.db_name.decode(),
            u"db_username": self.username.decode(),
            u"db_pwd": u"secret"
        }
        self.delete_instance_by_name(db_credentials, self.app_client)
        headers = {'Content-Type': 'application/json'}
        result = self.app_client.post(
            '/api/v2/postgraas_instances', headers=headers, data=json.dumps(db_credentials)
        )
        created_db = json.loads(result.data)
        created_db_id = created_db['postgraas_instance_id']
        actual = self.app_client.get(
            'api/v2/postgraas_instances/{}'.format(created_db_id), headers=headers
        )
        assert actual.status_code == 200
        actual_data = json.loads(actual.data)
        actual_data.pop('container_id')
        actual_data.pop('port')
        actual_data.pop('creation_timestamp')
        expected = {
            u'postgraas_instance_name': u'tests_postgraas_test_return_postgres_instance_api',
            u'db_name': self.db_name.decode(),
            u'username': self.username.decode(),
            u'password': u'',
            u'hostname': self.app_client.application.postgraas_backend.hostname,
            u'id': created_db_id,
        }
        assert actual_data == expected

        self.delete_instance_by_name(db_credentials, self.app_client)



import time

import psycopg2


def wait_for_postgres(dbname, user, password, host, port):
    """
    Try to connect to postgres every second, until it succeeds.
    """
    for i in range(540):
        try:
            psycopg2.connect(dbname=dbname, user=user, password=password, host=host, port=port)
            return
        except psycopg2.OperationalError:
            print i, " ..waiting for db"
            time.sleep(1)



import logging
from create_app import create_app
from configuration import get_config

logger = logging.getLogger(__name__)


app = create_app(get_config())


if __name__ == "__main__":
    app.run(debug=True, use_reloader=True)



import os
import pytest
from subprocess import check_call

import postgraas_server.backends.docker.postgres_instance_driver as pid
from postgraas_server import init_db
from postgraas_server.backends.docker import create_db

this_dir = os.path.dirname(__file__)


@pytest.yield_fixture
def cov_runner():
    # pretty hacky fixture to measure coverage of python scripts started as subprocess.
    # * run the script via 'coverage run', not via python
    # * move the created .coverage files two to ../.. so they are merged with the others
    check_call("rm -f .coverage.*", shell=True, cwd=this_dir)

    def run(script):
        command = "coverage run --branch --source ../.. -p {}".format(script)
        check_call(command, cwd=this_dir, shell=True)

    yield run
    check_call("mv .coverage.* ../..", shell=True, cwd=this_dir)


def test_create_init_db(cov_runner):
    # test the standalone 'create_db' and 'init_db' scripts.
    if pid.check_container_exists('postgraas_master_db'):
        pid.delete_postgres_instance('postgraas_master_db')
    create_db_file = create_db.__file__
    cov_runner(create_db_file)
    # also check the path where the container already exists:
    cov_runner(create_db_file)

    init_db_file = init_db.__file__
    cov_runner(init_db_file)



import datetime
import logging
import psycopg2

from flask import current_app
from flask_restful import fields, Resource, marshal_with, reqparse, abort
from flask_sqlalchemy import SQLAlchemy

from postgraas_server.backends.exceptions import PostgraasApiException

logger = logging.getLogger(__name__)

db = SQLAlchemy()


class DBInstance(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    postgraas_instance_name = db.Column(db.String(100))
    creation_timestamp = db.Column(db.DateTime)
    db_name = db.Column(db.String(100))
    username = db.Column(db.String(100))
    password = db.Column(db.String(100))
    hostname = db.Column(db.String(50))
    port = db.Column(db.Integer)
    container_id = db.Column(db.String(100))

    def __init__(
        self,
        postgraas_instance_name,
        db_name,
        username,
        password,
        hostname,
        port,
        container_id=None
    ):
        self.postgraas_instance_name = postgraas_instance_name
        self.creation_timestamp = datetime.datetime.now()
        self.db_name = db_name
        self.username = username
        self.password = password
        self.hostname = hostname
        self.port = port
        self.container_id = container_id

    def __repr__(self):
        return '<PGName {} User {} Docker_id {}>'.format(
            self.postgraas_instance_name, self.username, self.container_id
        )


db_instance_marshaller = {
    'id': fields.Integer,
    'postgraas_instance_name': fields.String,
    'creation_timestamp': fields.DateTime(dt_format='iso8601'),
    'db_name': fields.String,
    'username': fields.String,
    'password': fields.String,
    'hostname': fields.String,
    'port': fields.String,
    'container_id': fields.String,
}


class DBInstanceResource(Resource):
    @marshal_with(db_instance_marshaller)
    def get(self, id):
        entity = DBInstance.query.get(id)
        return entity

    def delete(self, id):
        parser = reqparse.RequestParser()
        parser.add_argument('db_pwd', required=True, type=str, help='pass of the db user')
        args = parser.parse_args()

        entity = DBInstance.query.get(id)
        if not entity:
            return {'status': 'failed', 'msg': 'Postgraas instance {} does not exist'.format(id)}

        connection_error = None
        try:
            conn = psycopg2.connect(
                user=entity.username,
                password=args['db_pwd'],
                host=current_app.postgraas_backend.master_hostname,
                port=entity.port,
                dbname=entity.db_name
            )
            conn.close()
        except StandardError as ex:
            connection_error = str(ex)

        if connection_error is not None:
            return {
                'status': 'failed',
                'msg': 'Could not connect to postgres instance: {}'.format(connection_error)
            }

        if not current_app.postgraas_backend.exists(entity):
            logger.warning(
                "container {} does not exist, how could that happen?".format(entity.container_id)
            )
            db.session.delete(entity)
            db.session.commit()
            return {
                'status': 'success',
                'msg': 'deleted postgraas instance, but container was not found...'
            }

        try:
            current_app.postgraas_backend.delete(entity)
        except PostgraasApiException as e:
            logger.warning("error deleting container {}: {}".format(entity.container_id, str(e)))
            return {'status': 'failed', 'msg': str(e)}
        db.session.delete(entity)
        db.session.commit()
        return {'status': 'success', 'msg': 'deleted postgraas instance'}


class DBInstanceCollectionResource(Resource):
    @marshal_with(db_instance_marshaller)
    def get(self):
        all = DBInstance.query.all()
        return all

    def post(self):
        parser = reqparse.RequestParser()
        parser.add_argument(
            'postgraas_instance_name',
            required=True,
            type=str,
            help='name of the postgraas instance'
        )
        parser.add_argument('db_name', required=True, type=str, help='name of the db')
        parser.add_argument('db_username', required=True, type=str, help='username of the db')
        parser.add_argument('db_pwd', required=True, type=str, help='pass of the db user')
        args = parser.parse_args()

        if DBInstance.query.filter_by(postgraas_instance_name=args['postgraas_instance_name']
                                      ).first():
            return {
                'msg':
                "postgraas_instance_name already exists {}".format(args['postgraas_instance_name'])
            }

        db_credentials = {
            "db_name": args['db_name'],
            "db_username": args['db_username'],
            "db_pwd": args['db_pwd'],
            "host": current_app.postgraas_backend.hostname,
            "port": current_app.postgraas_backend.port
        }

        db_entry = DBInstance(
            postgraas_instance_name=args['postgraas_instance_name'],
            db_name=args['db_name'],
            username=db_credentials['db_username'],
            password="",
            hostname=db_credentials['host'],
            port=db_credentials['port']
        )
        if current_app.postgraas_backend.exists(db_entry):
            abort(409,
                description="database or user already exists {}, {}".format(args['db_name'], args['db_username'])
            )

        try:
            db_entry.container_id = current_app.postgraas_backend.create(db_entry, db_credentials)
        except PostgraasApiException as e:
            return {'msg': str(e)}

        if '@' not in args['db_username']:
            try:
                username = '@'.join([args['db_username'], current_app.postgraas_backend.server])
            except (AttributeError, KeyError):
                username = args['db_username']
            db_entry.username = username

        db.session.add(db_entry)
        db.session.commit()
        db_credentials["container_id"] = db_entry.container_id
        db_credentials["postgraas_instance_id"] = db_entry.id
        db_credentials["db_username"] = username
        return db_credentials, 201



import psycopg2
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT


def _create_pg_connection(config):
    if 'server' in config:
        username = '@'.join([config['username'], config['server']])
    else:
        username = config['username']
    return psycopg2.connect(
        database=config['database'],
        user=username,
        host=config['host'],
        port=config['port'],
        password=config['password'],
    )


def check_db_or_user_exists(db_name, db_user, config):
    con = _create_pg_connection(config)
    cur = con.cursor()
    cur.execute("SELECT 1 FROM pg_database WHERE datname='{}';".format(db_name))
    db_exists = cur.fetchone() is not None
    cur.execute("SELECT 1 FROM pg_roles WHERE rolname='{}';".format(db_user))
    user = cur.fetchone()
    user_exists = user is not None
    return db_exists or user_exists


def create_postgres_db(connection_dict, config):
    if check_db_or_user_exists(connection_dict["db_name"], connection_dict["db_username"], config):
        raise ValueError("db or user already exists")
    con = _create_pg_connection(config)
    con.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
    cur = con.cursor()
    create_role = "CREATE USER {db_username} WITH PASSWORD '{db_pwd}';".format(**connection_dict)
    drop_role = "DROP ROLE {db_username};".format(**connection_dict)
    grant_role = 'GRANT {db_username} TO "{postgraas_user}";'.format(
        db_username=connection_dict['db_username'], postgraas_user=config['username'].split('@')[0]
    )
    create_database = "CREATE DATABASE {db_name} OWNER {db_username};".format(**connection_dict)
    try:
        cur.execute(create_role)
        cur.execute(grant_role)
    except psycopg2.ProgrammingError as e:
        raise ValueError(e.args[0])
    # cleanup role in case database creation fails
    # saidly 'CREATE DATABASE' cannot run inside a transaction block
    try:
        cur.execute(create_database)
    except psycopg2.ProgrammingError as e:
        cur.execute(drop_role)
        raise ValueError(e.args[0])


def delete_database(db_name, config):
    con = _create_pg_connection(config)
    con.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
    cur = con.cursor()
    try:
        cur.execute('''DROP DATABASE "{}";'''.format(db_name))
    except psycopg2.ProgrammingError as e:
        raise ValueError(e.args[0])



def delete_user(username, config):
    con = _create_pg_connection(config)
    con.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
    cur = con.cursor()
    try:
        cur.execute('''DROP USER "{}";'''.format(username))
    except psycopg2.ProgrammingError as e:
        raise ValueError(e.args[0])


class PostgraasApiException(Exception):
    """Base exception which should be used by backends instead of their own
    backend specific exceptions.

    .. example::

        try:
            docker_backend.do_something()
        except docker.errors.APIError as e:
            raise PostgraasApiException(str(e))

    """

    def __init__(self, message=None):
        self.message = message

    def __str__(self):
        return self.message



from flask import Flask
from flask_restful import Api

from postgraas_server.management_resources import DBInstanceResource, DBInstanceCollectionResource, db
from postgraas_server.backends import get_backend
from postgraas_server.configuration import get_application_config, get_meta_db_config_path


INT_OPTIONS = [
    'SQLALCHEMY_POOL_RECYCLE',
    'SQLALCHEMY_POOL_SIZE',
    'SQLALCHEMY_POOL_TIMEOUT',
    'SQLALCHEMY_MAX_OVERFLOW'
]


def create_app(config):
    app = Flask(__name__)
    app.config['SQLALCHEMY_DATABASE_URI'] = get_meta_db_config_path(config)
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
    app_config = get_application_config(config)
    for key, value in app_config:
        app.config[key.upper()] = int(value) if key.upper() in INT_OPTIONS else value
    restful_api = Api(app)
    restful_api.add_resource(DBInstanceResource, "/api/v2/postgraas_instances/<int:id>")
    restful_api.add_resource(DBInstanceCollectionResource, "/api/v2/postgraas_instances")
    db.init_app(app)
    app.postgraas_backend = get_backend(config)

    @app.route('/health')
    def health():
        return "ok"

    return app



import io
import os
import json
import errno
import logging
import ConfigParser

__all__ = ['get_config', 'get_default_config_filename', 'get_application_config', 'expand_env_vars']

logger = logging.getLogger(__name__)


def get_default_config_filename():
    config_filename = os.path.join(os.path.abspath(os.getcwd()), "application.cfg")
    return config_filename


def _load_secrets(filename='/secrets'):
    try:
        with open(filename, 'rb') as secrets_file:
            secrets = json.loads(secrets_file.read())
    except IOError as e:
        if e.errno in (errno.ENOENT, errno.EISDIR):
            return {}
        e.strerror = 'Unable to load configuration file (%s)' % e.strerror
        raise
    return secrets


def get_config(config_filename=get_default_config_filename(), secrets_file='/secrets'):
    config = ConfigParser.RawConfigParser()
    logger.debug('config filename: {}'.format(config_filename))
    secrets = _load_secrets(filename=secrets_file)
    if secrets:
        from cryptography.fernet import Fernet
        f = Fernet(secrets['encryption_key'].encode())
        with open(config_filename, 'rb') as cfg:
            cfg_content = f.decrypt(cfg.read())
        config.readfp(io.BytesIO(cfg_content))
    else:
        config.read(config_filename)
    expand_env_vars(config)
    return config


def expand_env_vars(config):
    for sec in config.sections():
        for opt in config.options(sec):
            config.set(sec, opt, os.path.expandvars(config.get(sec, opt)))


def get_application_config(config):
    try:
        return config.items('application')
    except ConfigParser.NoSectionError:
        return []


def get_meta_db_config_path(config):
    try:
        server = config.get('metadb', 'server')
        username = '@'.join([config.get('metadb', 'db_username'), server])
    except ConfigParser.NoOptionError:
        username = config.get('metadb', 'db_username')

    db_path = 'postgresql://{db_username}:{db_pwd}@{host}:{port}/{db_name}'.format(
        db_name=config.get('metadb', 'db_name'),
        db_username=username,
        db_pwd=config.get('metadb', 'db_pwd'),
        host=config.get('metadb', 'host'),
        port=config.get('metadb', 'port')
    )
    return db_path



import os
import unittest

import postgraas_server.configuration as configuration
from postgraas_server.create_app import create_app
from postgraas_server.management_database import is_sane_database


class TestManagementDatabase(unittest.TestCase):
    def setUp(self):
        self.module_path = os.path.abspath(os.path.dirname(__file__))
        self.this_app = create_app(
            configuration.get_config(os.path.join(self.module_path, 'application.cfg'))
        )
        self.this_app.config['SQLALCHEMY_DATABASE_URI'] = "sqlite://"
        self.this_app.use_reloader = False
        self.this_app.config['TESTING'] = True

    def tearDown(self):
        from postgraas_server.management_resources import db
        with self.this_app.app_context():
            db.drop_all()

    def test_is_sane_database(self):
        from postgraas_server.management_resources import db
        with self.this_app.app_context():
            db.drop_all()
            db.create_all()
            is_sane = is_sane_database(db.Model, db.session)

        self.assertEqual(True, is_sane)

    def test_is_not_sane_database(self):
        from postgraas_server.management_resources import db
        with self.this_app.app_context():
            is_sane = is_sane_database(db.Model, db.session)
        self.assertEqual(False, is_sane)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# coding=utf-8

import logging
from contextlib import contextmanager
import configuration as cfg
import datetime
import psycopg2
import click
from flask import Blueprint, Flask, Response, abort
from prometheus_client import CONTENT_TYPE_LATEST, generate_latest
from prometheus_client.core import REGISTRY, GaugeMetricFamily

logger = logging.getLogger(__name__)

_NO_TRACING = 0


@contextmanager
def db_connection(config):
    connection = psycopg2.connect(
        database=config.get('metadb', 'db_name'),
        user=config.get('metadb', 'db_username'),
        password=config.get('metadb', 'db_pwd'),
        host=config.get('metadb', 'host'),
        port=config.get('metadb', 'port')
    )
    yield connection
    connection.close()


def do_count_query(config):
    with db_connection(config) as conn:
        cursor = conn.cursor()
        sql = (
            "SELECT count(*) FROM pg_database "
            "WHERE datistemplate = false;"
        )
        cursor.execute(sql)
        (count,) = cursor.fetchone()

    return count


def count_postgraas_instances(config):
    count = do_count_query(config)
    return count


class CustomCollector(object):

    def __init__(self, config):
        self.metric_name = 'number_of_postgraas_instances'
        self.help = 'number of postgraas instances'
        self.config = config

    def collect(self):
        count = count_postgraas_instances(self.config)
        count_metric = GaugeMetricFamily(self.metric_name, self.help, value=count)
        yield count_metric

    def describe(self):
        yield GaugeMetricFamily(self.metric_name, self.help)

blueprint = Blueprint('monitoring', __name__)


@blueprint.route('/health')
def health():
    return 'ok'


@blueprint.route('/')
@blueprint.route('/metrics')
def metrics():
    try:
        content = generate_latest(REGISTRY)
        return content, 200, {'Content-Type': CONTENT_TYPE_LATEST}
    except Exception as error:
        logging.exception("Any exception occured during scraping")
        abort(Response("Scrape failed: {}".format(error), status=502))


@click.command()
@click.option("--test", is_flag=True, help="Print the results of a single readout instead of running a web server")
@click.option("--port", default=8000, help="Set the port for running the web server that is scraped by prometheus")
def run_server(test, port):
    """
    This script provides monitoring information concerning the export performed by the aedir-exporter CLI.
    """
    logging.basicConfig(level=logging.INFO, format="%(asctime)s %(name)-12s %(levelname)-8s %(message)s")

    config = cfg.get_config()

    collector = CustomCollector(config)
    if test:
        click.echo("TEST MODE")
        for metric in collector.collect():
            for sample in metric.samples:
                click.echo(sample)
    else:
        click.echo("Running web server at port {}".format(port))
        REGISTRY.register(collector)
        app = Flask(__name__)
        app.register_blueprint(blueprint)
        app.run(host='0.0.0.0', port=port, threaded=True)


if __name__ == '__main__':
    run_server()


import time

from postgraas_server.backends.docker.postgres_instance_driver import _docker_client


def wait_for_postgres_listening(container_id, timeout=10):
    """
    Inspect/follow the output of the docker container logs until encountering the
    message from postgres that it accepts connections.

    Raises in case the container does not exist.

    Caveat: this check is specific to the used docker image.
    Tested for current postgres:9.4 image.

    :returns: boolean, whether the message has been (True), or the timeout has been hit (False).
    """
    c = _docker_client()
    cont = c.containers.get(container_id)
    for i in range(max(int(timeout * 10), 1)):
        # not very efficient to always gather all logs, but as we only use it for testing
        # and it seems to be fast enough, should be OK.
        output = cont.logs(stdout=True, stderr=True)
        # the startup script in the docker image starts postgres twice,
        # so wait for the second start:
        if output.count('database system is ready to accept connections') >= 2:
            return True
        time.sleep(0.1)
    return False



import ConfigParser

from postgraas_server.configuration import get_config
from postgraas_server.management_database import init_db
from postgraas_server.utils import wait_for_postgres


def main():
    from postgraas_server import postgraas_api
    config = get_config()
    db_credentials = {
        "db_name": config.get('metadb', 'db_name'),
        "db_username": config.get('metadb', 'db_username'),
        "db_pwd": config.get('metadb', 'db_pwd'),
        "host": config.get('metadb', 'host'),
        "port": config.get('metadb', 'port')
    }

    try:
        server = config.get('metadb', 'server')
        db_credentials['db_username'] = '@'.join([db_credentials['db_username'], server])
    except ConfigParser.NoOptionError:
        pass

    wait_for_postgres(
        db_credentials['db_name'], db_credentials['db_username'], db_credentials['db_pwd'],
        db_credentials['host'], db_credentials['port']
    )
    print "initializing db"
    init_db(db_credentials, postgraas_api.app)


if __name__ == '__main__':
    main()



import logging
from sqlalchemy import inspect
from sqlalchemy.ext.declarative.clsregistry import _ModuleMarker
from sqlalchemy.orm import RelationshipProperty

logger = logging.getLogger(__name__)


def is_sane_database(Base, session):
    """
    from: http://stackoverflow.com/questions/30428639/check-database-schema-matches-sqlalchemy-models-on-application-startup
    Check whether the current database matches the models declared in model base.

    Currently we check that all tables exist with all columns. What is not checked

    * Column types are not verified

    * Relationships are not verified at all (TODO)

    :param Base: Declarative Base for SQLAlchemy models to check

    :param session: SQLAlchemy session bound to an engine

    :return: True if all declared models have corresponding tables and columns.
    """
    engine = session.get_bind()
    iengine = inspect(engine)
    errors = False
    tables = iengine.get_table_names()

    # Go through all SQLAlchemy models
    for name, klass in Base._decl_class_registry.items():
        if isinstance(klass, _ModuleMarker):
            # Not a model
            continue
        table = klass.__tablename__
        if table in tables:
            # Check all columns are found
            # Looks like [{'default': "nextval('sanity_check_test_id_seq'::regclass)", 'autoincrement': True, 'nullable': False, 'type': INTEGER(), 'name': 'id'}]
            columns = [c["name"] for c in iengine.get_columns(table)]
            mapper = inspect(klass)
            for column_prop in mapper.attrs:
                if isinstance(column_prop, RelationshipProperty):
                    # TODO: Add sanity checks for relations
                    pass
                else:
                    for column in column_prop.columns:
                        # Assume normal flat column
                        if column.key not in columns:
                            logger.error(
                                "Model %s declares column %s which does not exist in database %s",
                                klass, column.key, engine
                            )
                            errors = True
        else:
            logger.error(
                "Model %s declares table %s which does not exist in database %s", klass, table,
                engine
            )
            errors = True
    return not errors


def init_db(db_credentials, postgraas_app):
    # import all modules here that might define models so that
    # they will be registered properly on the metadata.  Otherwise
    # you will have to import them first before calling init_db()

    from postgraas_server.management_resources import db
    with postgraas_app.app_context():
        if not is_sane_database(db.Model, db.session):
            db.drop_all()
            db.create_all()



import os
import uuid
import json
import ConfigParser
import StringIO
import pytest

import postgraas_server.configuration as configuration
from postgraas_server.create_app import create_app
import postgraas_server.backends.postgres_cluster.postgres_cluster_driver as pgcd
from postgraas_server.backends.postgres_cluster import PGClusterBackend

CLUSTER_CONFIG = """[metadb]
db_name = postgraas
db_username = postgraas
db_pwd = postgraas12
host = localhost
port = 54321

[backend]
type = pg_cluster
host = {host}
port = {port}
database = {database}
username = {username}
password = {password}
""".format(
    database=os.environ.get('PGDATABASE', 'postgres'),
    username=os.environ.get('PGUSER', 'postgres'),
    password=os.environ.get('PGPASSWORD', 'postgres'),
    port=os.environ.get('PGPORT', '5432'),
    host=os.environ.get('PGHOST', 'localhost')
)

CONFIGS = {
    'pg_cluster': CLUSTER_CONFIG,
}


def remove_digits(s):
    return ''.join(c for c in s if not c.isdigit())


def delete_test_database_and_user(db_name, username, config):
    try:
        pgcd.delete_database(db_name, config)
    except ValueError as e:
        pass
        #print ValueError(e.args[0])
    try:
        pgcd.delete_user(username, config)
    except ValueError as e:
        pass
        #print ValueError(e.args[0])

@pytest.fixture(params=['pg_cluster'])
def parametrized_setup(request, tmpdir):
    from postgraas_server.management_resources import db
    cfg = tmpdir.join('config')
    cfg.write(CONFIGS[request.param])
    config = configuration.get_config(cfg.strpath)
    this_app = create_app(config)
    this_app.config['SQLALCHEMY_DATABASE_URI'] = "sqlite://"
    this_app.use_reloader = False
    this_app.config['TESTING'] = True
    ctx = this_app.app_context()
    ctx.push()
    db.create_all()
    username, db_name = str(uuid.uuid4()).replace('-', '_'), str(uuid.uuid4()).replace('-', '_')
    request.cls.app_client = this_app.test_client()
    request.cls.db_name = remove_digits(db_name)
    request.cls.username = remove_digits(username)
    request.cls.backend = request.param
    yield
    if request.param == 'pg_cluster':
        try:
            delete_test_database_and_user(db_name, username, dict(config.items('backend')))
        except Exception:
            pass
    db.drop_all()
    ctx.pop()

class PostgraasApiTestBase:
    def get_postgraas_by_name(self, name, client):
        headers = {'Content-Type': 'application/json'}
        list = client.get('/api/v2/postgraas_instances', headers=headers)
        for instance in json.loads(list.data):
            if instance["postgraas_instance_name"] == name:
                return instance["id"]

    def delete_instance_by_name(self, db_credentials, client):
        id = self.get_postgraas_by_name(db_credentials["postgraas_instance_name"], client)
        db_pwd = db_credentials["db_pwd"]
        headers = {'Content-Type': 'application/json'}
        client.delete(
            '/api/v2/postgraas_instances/' + str(id),
            data=json.dumps({
                'db_pwd': db_pwd
            }),
            headers=headers
        )

@pytest.mark.usefixtures('parametrized_setup')
class TestPostgraasApi(PostgraasApiTestBase):

    def test_delete_db_and_user(self):
        config = ConfigParser.ConfigParser()

        config.readfp(StringIO.StringIO(CONFIGS[self.backend]))
        backend_config = dict(config.items('backend'))
        db_credentials = {
            "postgraas_instance_name": "tests_postgraas_test_create_postgres_instance_api",
            "db_name": 'test_create_postgres_instance_exists',
            "db_username": 'test_create_postgres_instance_exists_username',
            "db_pwd": 'test_db_pwd',
            "host": backend_config['host'],
            "port": backend_config['port']
        }
        delete_test_database_and_user(db_credentials['db_name'], db_credentials['db_username'], backend_config)
        exists = pgcd.check_db_or_user_exists(db_credentials['db_name'], db_credentials['db_username'], backend_config)
        assert exists is False
        pgcd.create_postgres_db(db_credentials, backend_config)
        exists = pgcd.check_db_or_user_exists(db_credentials['db_name'], db_credentials['db_username'], backend_config)
        assert exists is True
        response = self.app_client.post('/api/v2/postgraas_instances',
                                        data=json.dumps(db_credentials),
                                        headers={'Content-Type': 'application/json'})
        print response.data
        assert ("database or user already exists" in json.loads(response.data)['description']) is True
        delete_test_database_and_user(db_credentials['db_name'], db_credentials['db_username'], backend_config)
        response = self.app_client.post('/api/v2/postgraas_instances',
                                        data=json.dumps(db_credentials),
                                        headers={'Content-Type': 'application/json'})
        print response.data
        assert ("test_create_postgres_instance_exists" in json.loads(response.data)['db_name']) is True


    def test_create_postgres_instance_exists(self):
        config = ConfigParser.ConfigParser()

        config.readfp(StringIO.StringIO(CONFIGS[self.backend]))
        backend_config = dict(config.items('backend'))
        print config
        db_credentials = {
            "postgraas_instance_name": "tests_postgraas_test_create_postgres_instance_api",
            "db_name": 'test_create_postgres_instance_exists',
            "db_username": 'test_create_postgres_instance_exists_username',
            "db_pwd": 'test_db_pwd',
            "host": backend_config['host'],
            "port": backend_config['port']
        }
        delete_test_database_and_user(db_credentials['db_name'], db_credentials['db_username'], backend_config)
        exists = pgcd.check_db_or_user_exists(db_credentials['db_name'], db_credentials['db_username'], backend_config)
        assert exists is False
        print db_credentials
        pgcd.create_postgres_db(db_credentials, backend_config)
        exists = pgcd.check_db_or_user_exists(db_credentials['db_name'], db_credentials['db_username'], backend_config)
        assert exists is True
        response = self.app_client.post('/api/v2/postgraas_instances',
                                        data=json.dumps(db_credentials),
                                        headers={'Content-Type': 'application/json'})
        delete_test_database_and_user(db_credentials['db_name'], db_credentials['db_username'], backend_config)
        assert ("database or user already exists" in json.loads(response.data)['description']) is True

    def test_create_postgres_instance_username_exists(self):
        config = ConfigParser.ConfigParser()

        config.readfp(StringIO.StringIO(CONFIGS[self.backend]))
        backend_config = dict(config.items('backend'))
        print config
        db_credentials = {
            "postgraas_instance_name": "tests_postgraas_test_create_postgres_instance_api",
            "db_name": 'test_create_postgres_instance_exists',
            "db_username": 'test_create_postgres_instance_exists_username',
            "db_pwd": 'test_db_pwd',
            "host": backend_config['host'],
            "port": backend_config['port']
        }
        db_credentials_same_user = {
            "postgraas_instance_name": "tests_postgraas_test_create_postgres_instance_api",
            "db_name": 'test_create_postgres_instance_exists_new_db_name',
            "db_username": 'test_create_postgres_instance_exists_username',
            "db_pwd": 'test_db_pwd',
            "host": backend_config['host'],
            "port": backend_config['port']
        }
        delete_test_database_and_user(db_credentials['db_name'], db_credentials['db_username'], backend_config)
        exists = pgcd.check_db_or_user_exists(db_credentials['db_name'], db_credentials['db_username'], backend_config)
        assert exists is False
        print db_credentials
        pgcd.create_postgres_db(db_credentials, backend_config)
        exists = pgcd.check_db_or_user_exists(db_credentials['db_name'], db_credentials['db_username'], backend_config)
        assert exists is True
        with pytest.raises(ValueError):
            pgcd.create_postgres_db(db_credentials_same_user, backend_config)

        response = self.app_client.post('/api/v2/postgraas_instances',
                                        data=json.dumps(db_credentials_same_user),
                                        headers={'Content-Type': 'application/json'})
        delete_test_database_and_user(db_credentials['db_name'], db_credentials['db_username'], backend_config)
        delete_test_database_and_user(db_credentials_same_user['db_name'], db_credentials_same_user['db_username'], backend_config)
        assert ("database or user already exists" in json.loads(response.data)['description']) is True


from . import postgres_cluster_driver as pgcd


class PGClusterBackend(object):
    def __init__(self, config):
        self.config = config

    def create(self, entity, connection_info):
        pgcd.create_postgres_db(connection_info, self.config)
        return entity.id

    def delete(self, entity):
        pgcd.delete_database(entity.db_name, self.config)
        pgcd.delete_user(entity.username, self.config)

    def exists(self, entity):
        return pgcd.check_db_or_user_exists(entity.db_name, entity.username, self.config)

    @property
    def hostname(self):
        return self.config['host']

    @property
    def port(self):
        return self.config['port']

    @property
    def server(self):
        return self.config['server']

    @property
    def master_hostname(self):
        return self.hostname



from postgraas_server.backends.docker import postgres_instance_driver as pg
from postgraas_server.configuration import get_config
from postgraas_server.utils import wait_for_postgres


def create_db_container():
    config = get_config()
    db_credentials = {
        "db_name": config.get('metadb', 'db_name'),
        "db_username": config.get('metadb', 'db_username'),
        "db_pwd": config.get('metadb', 'db_pwd'),
        "host": config.get('metadb', 'host'),
        "port": config.get('metadb', 'port')
    }
    if pg.check_container_exists('postgraas_master_db'):
        print "warning container already exists"
        postgraas_db = pg.get_container_by_name('postgraas_master_db')
        db_credentials['container_id'] = postgraas_db.id
    else:
        db_credentials['container_id'] = pg.create_postgres_instance(
            'postgraas_master_db', db_credentials
        )
    return db_credentials


def main():
    print "creating container for the management db"
    db_credentials = create_db_container()
    wait_for_postgres(
        db_credentials['db_name'], db_credentials['db_username'], db_credentials['db_pwd'],
        db_credentials['host'], db_credentials['port']
    )


if __name__ == '__main__':
    main()



#from . import postgres_instance_driver as pg
from ..exceptions import PostgraasApiException


class DockerBackend(object):
    def __init__(self, config=None):
        self.config = config

    def create(self, entity, connection_info):
        from docker.errors import APIError
        from . import postgres_instance_driver as pg
        try:
            return pg.create_postgres_instance(entity.postgraas_instance_name, connection_info)
        except APIError as e:
            raise PostgraasApiException(str(e))

    def delete(self, entity):
        from docker.errors import APIError
        from . import postgres_instance_driver as pg
        try:
            return pg.delete_postgres_instance(entity.container_id)
        except APIError as e:
            raise PostgraasApiException(str(e))

    def exists(self, entity):
        from . import postgres_instance_driver as pg
        if entity.postgraas_instance_name:
            return pg.check_container_exists(entity.postgraas_instance_name)
        else:
            False

    @property
    def hostname(self):
        from . import postgres_instance_driver as pg
        return pg.get_hostname()

    @property
    def port(self):
        from . import postgres_instance_driver as pg
        return pg.get_open_port()

    @property
    def master_hostname(self):
        return '127.0.0.1'
